"""
This module provides various functions and classes to manage a version control system similar to Git.
It includes functionalities for creating commits, branches, tags, merging branches, and more.
Functions:
    get_merge_base(o1, o2):
        Returns the commit id of the common ancestor of two commits.
    delete_ref(name):
        Deletes the reference by its name.
    get_ref_path(ref):
        Returns the path of the reference using its name.
    merge(other, head):
        Merges the specified branch into the head branch.
    status():
        Prints the current branch we are on.
    reset(o_id):
        Resets the head and branch to a particular commit.
    show(o_id):
    get_all_branches():
        Prints the names of all branches, along with the current branch.
    tag(tag_name, o_id):
        Tags a commit with a specified name.
    checkout(ref_name=None, o_id=None):
        Checks out to a commit or reference.
    branch(name, o_id):
        Creates a branch with the specified name and commit id.
    log_(o_id=None, ref_name=None):
        Prints the chain of commits with respect to the current commit.
    commit(msg="No message left"):
        Creates a commit with the specified message and returns the commit id.
    get_commit(o_id):
        Returns the commit object for a particular commit id.
    iter_commits_and_parents(oid):
        Yields the commit id of the commit and all its parent commits.
    snapshot(directory="."):
        Creates a snapshot for the current directory structure and returns the object id of this directory.
    read_snapshot(o_id, base_path='.'):
        Writes the current file structure as per the snapshot.
Classes:
    Commit:
        Represents a commit object with attributes like snapshot, author, time, message, and parents.
"""


from . import data
from . import diffs
import os, shutil
import zlib
import datetime
import collections


def get_merge_base(o1, o2):
    '''
    Returns the commit id of the common ancestor.
    '''
    parents1 = set(iter_commits_and_parents(o1))

    for o in iter_commits_and_parents(o2):
        if o in parents1:
            return o

def delete_ref(name):
    '''
    Deletes the reference.
    '''
    path = get_ref_path(name)
    os.remove(path)

def get_ref_path(ref):
    '''
    Returns the path of the reference using it's name.
    '''
    potential_paths = [f"{ref}", f"refs\\heads\\{ref}", f"refs\\tags\\{ref}"]
    for path in potential_paths:
        if os.path.exists(f"{data.GIT_DIR}\\{path}"):
            return f"{data.GIT_DIR}\\{path}"
    return None

def merge(other, head):
    '''
    Merges the branch into head.
    '''
    other = data.get_ref(f'refs\\heads\\{other}')
    head = data.get_ref(f'refs\\heads\\{head}')
    
    other_commit = get_commit(other).snap
    head_commit = get_commit(head).snap

    diffs.merge_snaps(other_commit, head_commit)

    data.set_ref('MERGE_HEAD', other)
    print("Merged! Please make a commit to record the changes!")


def status():
    '''
    Prints the current branch we are on (latter on we will update it to even display the unstaged files).
    '''
    if data.check_symbolic():
        print(f"On branch {os.path.split(data.get_head_branch())[1]}")
    else:
        print(f"HEAD detached at {data.get_ref('HEAD')}")


def reset(o_id):
    '''
    This is different than checkout as in this case, the head as well as the branch would be set to a particular commit.
    This is not the case with checkout as checkout will just point the head to a commit.
    '''
    if data.check_symbolic():
        data.set_ref(data.get_head_branch(), o_id)
    else:
        data.set_head(o_id=o_id)
    c = get_commit(o_id=o_id)
    read_snapshot(c.snap)


def show(o_id):
    '''
    Prints the current commit details and diffs between the commit and its parent commit.
    '''
    cmt = get_commit(o_id)
    print(f"commit {o_id}\nauthor: {cmt.author}\ndate: {cmt.time}\nmessage: {cmt.msg}\n")
    p = cmt.parents
    # diffs.diff(cmt.parents[0], o_id)  # from, to
    
    if p:
        diffs.diff(p[0], o_id)
    else:
        print("This is the initial commit - no parent to compare with.")


def get_all_branches():
    '''
    Prints the names of all branches, along with with what branch we are pointing to right now
    '''
    head = None
    if data.check_symbolic():
        head = os.path.split(data.get_head_branch())[-1]

    for branch in os.listdir(os.path.join(data.GIT_DIR, 'refs', 'heads')):
        if head and branch == head:
            print("* " + branch)
        else:
            print("  " + branch)


def tag(tag_name, o_id):
    '''
    This is used to tag a commit
    '''
    data.set_ref(os.path.join('refs', 'tags', tag_name), o_id)


def checkout(ref_name = None, o_id = None):
    '''
    To checkout to a commit. Again this only changes the head position, branch will be unaffected.
    '''
    if ref_name:
        for ref, id in data.iter_refs():
            if not id: continue
            if os.path.split(ref)[-1] == ref_name:
                o_id = id
                if ref.split("\\")[1] == 'heads':
                    data.set_head(ref = ref)
                else:
                    data.set_head(o_id = o_id)
                try:
                    c = get_commit(o_id)
                    print(o_id)
                    read_snapshot(c.snap)
                except: print("Invalid reference")
                return

        if not o_id:
            print("This tag doesn't point to any commit!")
            return

    c = get_commit(o_id)
    read_snapshot(c.snap)
    data.set_head(o_id = o_id)

def branch(name, o_id):
    '''
    Creates a branch.
    '''
    data.set_ref(os.path.join('refs', 'heads', name), o_id)


def log_(o_id = None, ref_name = None):
    '''
    Prints the chain of commits wrt current commit (it and all its parent commits).
    '''
    if not o_id:
        if ref_name:
            o_id = data.get_ref(f'refs\\tags\\{ref_name}')
            if o_id is None:
                o_id = data.get_ref(f'refs\\heads\\{ref_name}')
            
            if o_id is None:
                print("This reference doesn't exist!")
                return
        else:
            o_id = data.get_ref('HEAD')
    commit = data.get_object(o_id, expected='commit').decode().split("\n")
    
    refs = {}
    for refname, refv in data.iter_refs():
        if not refv: continue
        refs.setdefault(refv, []).append(os.path.split(refname)[-1])
    while len(commit) > 4:
        parent_o_id = commit[1].split(' ')[1]
        if refs.get(o_id):
            print(", ".join(refs[o_id]))
        print(f"commit {o_id}\n" + "\n".join(commit), "\n")

        commit = data.get_object(parent_o_id, 'commit').decode().split("\n")
        o_id = parent_o_id
    print(f"commit {o_id}\n" + "\n".join(commit), "\n")


class Commit:
    def __init__(self, snap, author, time, msg, parents = None):
        self.snap = snap
        self.parents = parents
        self.author = author
        self.time = time
        self.msg = msg


def commit(msg = "No message left"):
    '''
    Creates a commit.
    Returns the object id of commit.
    '''

    commit = f"snap {snapshot()}\n"

    head = data.get_ref('HEAD')                                 # retrieve the parent commit of this
    if head:
        commit += f"parent {head}\n"                            # attach the parent head to the commit
    
    merged_head = data.get_ref('MERGE_HEAD')
    if merged_head:
            commit += f"parent {merged_head}\n"
            delete_ref('MERGE_HEAD')

    commit += f"author x\ntime {datetime.datetime.now()}\nmessage {msg}"

    o_id = data.hash_object(commit.encode(), 'commit')          # generate the o_id for this commit
    
    branch = data.get_head_branch()

    print(f"[{os.path.split(branch)[-1]}] {msg}")
    data.set_ref(branch, o_id)                                  # set the current commit as the head, as this is the latest commit

    # now this has become like a linked list, with links as parent_o_id
    return o_id

def get_commit(o_id):
    '''
    Returns the commit wrt to a particular commit id.
    '''
    commit = iter(data.get_object(o_id, 'commit').decode().splitlines())
    attr = {'parents':[]}
    for c in commit:
        x = c.split(" ", 1)
        if x[0] == 'parent':
            attr['parents'].append(x[1])
        else:
            attr[x[0]] = x[1]

    return Commit(
        snap = attr['snap'],
        author = attr['author'],
        time = attr['time'],
        msg = attr['message'],
        parents = attr['parents'],
    )

def iter_commits_and_parents(oid):
    '''
    Gets the commit as well as all of its parental commits. Yeilds commit ID.
    '''

    oids = collections.deque({oid})
    visited = set()

    while oids:
        oid = oids.popleft()
        if not oid or oid in visited:
            continue
        visited.add(oid)
        yield oid

        commit = get_commit(oid)
        oids.extendleft(commit.parents[:1])
        oids.extend(commit.parents[1:])


def snapshot(directory = "."):
    '''
    Creates a snapshot for the current directory structure and get the object id of this directory.
    '''
    entries = []
    ignore = set()
    if os.path.exists(".trackitignore"):
        with open(".trackitignore", 'r') as f:
            ignore = set(f.read().split("\n"))

    with os.scandir(directory) as itr:
        for i in itr:
            curr = os.path.join(directory, i.name)
            
            if i.name == '.trackit' or i.name in ignore: continue

            if i.is_dir(follow_symlinks=False):
                object_type = 'snap'
                o_id = snapshot(curr)                     # this is a recursive algo for getting the o_id of the this dir
            else:
                object_type = 'blob'
                with open(curr, 'rb') as f:
                    o_id = data.hash_object(f.read())

            entries.append((object_type, o_id, i.name))

    snap = "".join(f"{object_type} {oid} {name}\n" for object_type, oid, name in entries)
    
    return data.hash_object(snap.encode(), 'snap')

def read_snapshot(o_id, base_path = '.'):
    '''
    Writes the current file structure as per the snapshot.
    '''
    obj_path = os.path.join(data.GIT_DIR, 'objects', o_id[:2], o_id[2:])

    with open(obj_path, 'rb') as f:
        r = f.read()

    content = zlib.decompress(r).split(b'\0', 1)[1].decode()

    del_list = {i:1 for i in os.listdir(base_path) if i != '.trackit'}

    # if os.path.exists(".trackitignore"):
    try:
        with open(".trackitignore", 'r') as f:
            ignore = f.readlines()
        for i in ignore:
            del_list.pop(i, None)
    except FileNotFoundError: pass
    
    for item in content.split("\n"):
        if not item: continue
        item_type, o_id, name = item.split(" ", 2)
        curr_path = os.path.join(base_path, name)

        try:
            del_list.pop(name)
        except KeyError:
            pass


        if item_type == 'snap':
            try:
                os.mkdir(curr_path)
            except: pass
            read_snapshot(o_id, base_path = curr_path)
        else:
            with open(curr_path, 'w') as f:
                f.write(data.get_object(o_id).decode())
    
    for i in del_list:
        p = os.path.join(base_path, i)
        if os.path.isfile(p):
            os.remove(p)
        else:
            shutil.rmtree(p)


"""
This module provides various functions to perform diff and merging
Functions:
    compare_snaps(s1, s2, path="."):
        Gets the overall comparison of files in snap1 and snap2.
    diff(f, t):
        Prints the changes when going from commit f to commit t.
    merge_snaps(f, t):
        Merge commit f_o_id to commit t_o_id.
    merge_blobs(branch, head):
        Merges two files. If conflicts arise, then will be indicated.
"""


from . import base
from . import data
import difflib
import os
from collections import defaultdict


def compare_snaps(s1, s2, path = "."):
    '''
    Gets the overall comparison of files in snap1 and snap2.\n
    Returns {filename : {0: oid1, 1: oid2}}.
    '''
    objects = defaultdict(lambda: [None, None])
    snaps = defaultdict(lambda: [None, None])

    for i, s in enumerate([s1, s2]):
        snap = data.get_object(s, expected='snap').decode().split("\n")
        for obj in snap[:-1]:
            o = obj.split(" ", 2)
            type = o[0]
            o_id = o[1]
            if type == 'snap':
                sub_path = f"{path}\{o[2]}"
                snaps[sub_path][i] = o_id
                sub_objects, sub_snaps = compare_snaps(o_id, o_id, sub_path)

                for sub_name, sub_obj in sub_objects.items():
                    objects[sub_name][i] = sub_obj[i]
                for sub_name, sub_snap in sub_snaps.items():
                    objects[sub_snap][i] = sub_obj[i]
                continue

            name = f"{path}\{o[2]}"
            objects[name][i] = o_id
        
    return objects, snaps

def diff(f, t):
    '''
    Prints the changes when going from commit f to commit t.
    '''
    f = base.get_commit(f)
    t = base.get_commit(t)
    files, snaps = compare_snaps(f.snap, t.snap)

    for filename, o_ids in files.items():
        print(filename)
        if o_ids[0] == o_ids[1]: continue
        elif o_ids[0] and not o_ids[1]:
            print(f"{filename} was deleted!")
        elif not o_ids[0] and o_ids[1]:
            print(f"{filename} was created!")
        else:
            # this means the file was modified
            print("Changes in", filename)
            file1 = data.get_object(o_ids[0], expected='blob').decode().split("\r")
            file2 = data.get_object(o_ids[1], expected='blob').decode().split("\r")
            d = difflib.unified_diff(file1, file2, lineterm='')
        
            for line in d:
                print(line)
        print()

    for snap, o_ids in snaps.items():
        # print(snap)
        if o_ids[0] and not o_ids[1]:
            print(f"{snap} was deleted!")
        elif not o_ids[0] and o_ids[1]:
            print(f"{snap} was created!")
        print()
    
    # just in case
    return files


def merge_snaps(f, t):
    '''
    Merge commit other snapshot to head snapshot.\n
    It will update the current snapshot to merged snapshot.\n
    Parameters: branch, head
    '''

    files, snaps = compare_snaps(f, t)
    
    # writing the merged to current snapshot and creating a new one and committing it.
    for snap, o_ids in snaps.items():
        # print(filename)
        if o_ids[0] and not o_ids[1]:
            os.makedirs(snap, exist_ok=True)
            print(f"{snap} was created!")
        elif not o_ids[0] and o_ids[1]:
            if os.path.exists(snap):
                os.rmdir(snap)
                print(f"{snap} was deleted!")

    for filename, o_ids in files.items():
        print(filename)
        if o_ids[0] and o_ids[1]:
            merged = merge_blobs(o_ids[0], o_ids[1])
            print(merged)
            with open(filename, 'w') as f:
                f.write("\n".join(merged))
        elif o_ids[0]:
            print("created")
            with open(filename, 'w') as f:
                f.write(data.get_object(o_ids[0], expected='blob').decode())

        elif o_ids[1]:
            print("deleted")
            os.remove(filename)
        print()


def merge_blobs(branch, head):
    '''
    Merges two files. If conflicts arises, then will be indicated.\n
    Returns the merged file.
    '''
    branch = data.get_object(branch, expected='blob').decode().split("\r")
    head = data.get_object(head, expected='blob').decode().split("\r")

    diff = difflib.ndiff(head, branch)
    
    flag1 = True
    flag2 = True
    output = []
    conflict_occur = False

    for d in diff:
        if d.startswith('-'):   # conflict started
            if flag1:
                conflict_occur = True
                output.append("<<<<<<< HEAD")
                flag1 = False
            output.append(d[2:])

        elif d.startswith('+'):
            if flag1 == False:   # if flag1 is false then it has entered the conflict block
                if flag2:   # if flag2 is true then we are just starting out with the second part of conflict block
                    output.append("=======")
                    flag2 = False
            output.append(d[2:])

        else:
            if flag1 == False:  # if the flag1 is false then we are still in confliuct block
                if flag2:       # if the flag2 is True then we have not started the branch's conflict block
                    output.append("=======")
                output.append(">>>>>>> branch")
                flag1 = True
                flag2 = True
                # we have ended the conflict block here
            output.append(d[2:])
    # just in case if there are no items in the list, we need to close the conflict
    if flag1 == False:
        if flag2:
            output.append("=======")
        output.append(">>>>>>> branch")
        flag1 = True
        flag2 = True

    if conflict_occur: print("Conflict Occured!")
    return output



"""
This module provides a command-line interface (CLI) for the Trackit version control system. It uses the argparse library to parse command-line arguments and execute corresponding functions. The module supports various commands such as initializing a repository, committing changes, creating branches, tagging, and more.
Functions:
- init(args): Initializes a new Trackit repository.
- parse_args(): Parses command-line arguments and sets up subparsers for different commands.
- snapshot(arg): Creates a snapshot of the current state of the repository.
- read_snapshot(arg): Restores the repository to a previously stored snapshot.
- commit(arg): Commits changes to the repository with a message.
- log_(args): Displays the commit log.
- checkout(args): Checks out a specific commit or reference.
- tag(args): Tags a specific commit with a name.
- status(args): Displays the current status of the repository.
- branch(args): Manages branches in the repository.
- reset(args): Resets the repository to a specific commit.
- show(args): Shows the content of a specific commit.
- diff(args): Displays the differences between two commits.
- merge(args): Merges two branches.
- merge_base(args): Finds the common ancestor of two commits.
- hash_object(arg): Hashes a file and stores it as an object in the repository.
- read_object(arg): Retrieves the content of an object using its hash.
The main() function is the entry point of the CLI, which parses the arguments and calls the appropriate function based on the command provided.
"""

import argparse
from . import data
from . import base
from . import diffs
import os, sys


def init(args):
    data.init()
    print(f"Initialised an empty trackit repository at {os.getcwd()}\\{data.GIT_DIR}.")

def parse_args():
    parser = argparse.ArgumentParser()
    commands = parser.add_subparsers(dest='cmd')
    commands.required = True

    init_parser = commands.add_parser('init')
    init_parser.set_defaults(func = init)   # when 'init' is parsed, init function will be triggered

    hash_file_parser = commands.add_parser('hash-file')
    hash_file_parser.set_defaults(func = hash_object)
    hash_file_parser.add_argument('obj')

    cat_file_parser = commands.add_parser('read-file')
    cat_file_parser.set_defaults(func = read_object)
    cat_file_parser.add_argument('obj')     # extra arg to be stored in obj variable

    write_tree_parser = commands.add_parser('snapshot')
    write_tree_parser.set_defaults(func = snapshot)

    read_tree_parser = commands.add_parser('read-snapshot')
    read_tree_parser.set_defaults(func = read_snapshot)
    read_tree_parser.add_argument('snap')

    commit_parser = commands.add_parser('commit')
    commit_parser.set_defaults(func = commit)
    commit_parser.add_argument('-m', '--message', required=True)

    log_parser = commands.add_parser('log')
    log_parser.set_defaults(func = log_)
    log_parser.add_argument('o_id', nargs='?')
    log_parser.add_argument('-r', '--ref', required=False)

    checkout_parser = commands.add_parser('checkout')
    checkout_parser.set_defaults(func = checkout)
    checkout_parser.add_argument('o_id', nargs='?')
    checkout_parser.add_argument('-r', '--ref', required=False)

    tag_parser = commands.add_parser('tag')
    tag_parser.set_defaults(func = tag)
    tag_parser.add_argument('tag_name')
    tag_parser.add_argument('o_id', nargs='?')

    status_parser = commands.add_parser ('status')
    status_parser.set_defaults(func=status)

    branch_parser = commands.add_parser('branch')
    branch_parser.set_defaults(func=branch)
    branch_parser.add_argument('name', nargs='?')
    branch_parser.add_argument('o_id', nargs='?')

    reset_parser = commands.add_parser("reset")
    reset_parser.set_defaults(func = reset)
    reset_parser.add_argument("o_id")

    show_parser = commands.add_parser("show")
    show_parser.set_defaults(func = show)
    show_parser.add_argument("o_id", nargs='?')

    diff_parser = commands.add_parser("diff")
    diff_parser.set_defaults(func = diff)
    diff_parser.add_argument("f_o_id", nargs='?')
    diff_parser.add_argument("t_o_id", nargs='?')

    merge_parser = commands.add_parser("merge")
    merge_parser.set_defaults(func = merge)
    merge_parser.add_argument("other_branch")
    merge_parser.add_argument("head_branch", default='main', nargs='?')

    merge_base_parser = commands.add_parser ('merge-base')
    merge_base_parser.set_defaults (func=merge_base)
    merge_base_parser.add_argument('o1')
    merge_base_parser.add_argument('o2')

    return parser.parse_args()

def main():
    args = parse_args()
    args.func(args)


''' ************************************************************************* '''

def snapshot(arg):
    print(base.snapshot())    # this will print o_id of the snap

def read_snapshot(arg):
    base.read_snapshot(arg.snap)    # this will restore the repo to a previously stored instance/snap using the given o_id

def commit(arg):
    print(base.commit(arg.message))

def log_(args):
    base.log_(args.o_id, args.ref)      # this ref can be either head or tag

def checkout(args):
    if args.o_id is None:
        if args.ref:
            base.checkout(ref_name = args.ref)
            return
        else:
            args.o_id = data.get_ref('HEAD')
    base.checkout(o_id = args.o_id)

def tag(args):
    base.tag(args.tag_name, args.o_id)

def status(args):
    base.status()

def branch(args):
    if not args.name and not args.o_id:
        base.get_all_branches()
        return
    
    if not args.o_id:
        args.o_id = data.get_ref('HEAD')

    base.branch(args.name, args.o_id)

def reset(args):
    base.reset(args.o_id)

def show(args):
    if not args.o_id: args.o_id = data.get_ref('HEAD')
    base.show(args.o_id)

def diff(args):
    if not args.f_o_id:
        c = base.get_commit(args.t_o_id)
        
        # Assuming c.parents is a list of parent commit ids
        if not hasattr(c, 'parents') or not c.parents:
            print("This commit does not have any parent!")
            return
        args.f_o_id = c.parents[0]
    diffs.diff(args.f_o_id, args.t_o_id)

def merge(args):
    base.merge(args.other_branch, args.head_branch)

def merge_base(args):
    print(base.get_merge_base(args.o1, args.o2))

def hash_object(arg):   # this is used to store the object and reference it with an o_id
    with open(arg.obj, 'rb') as f:
        content = f.read()
    o_id = data.hash_object(content)  # content is in binary form
    print(o_id)

def read_object(arg):    # this is used to retrieve the object content usign o_id
    sys.stdout.flush()
    sys.stdout.buffer.write(data.get_object(arg.obj, expected = None))


"""
This module provides the core data-level functionalities for TrackIt, a basic version control system inspired by Git. It includes functions to initialize a repository, hash and store objects, retrieve objects, manage references, and handle the HEAD pointer.
Functions:
- init(): Generates the required file directory structure for TrackIt.
- hash_object(content, obj_type="blob"): Hashes the object content and stores it in the .trackit/objects directory. Returns the object ID.
- get_object(o_id, expected='blob'): Retrieves the object content using the object ID. Returns the content in binary format.
- set_head(ref=None, o_id=None): Sets the HEAD to a specific reference or commit ID.
- get_head_branch(): Gets the branch or commit ID to which the HEAD is pointing.
- set_ref(ref, o_id): Sets or updates a reference to a specific commit ID.
- get_ref(ref): Gets the commit ID of a reference.
- iter_refs(): Iterates over all references (branches and tags) in the current TrackIt repository. Yields ref and object ID.
- check_symbolic(): Checks if the current HEAD is detached.
Constants:
- GIT_DIR: The directory name for TrackIt repository metadata.
"""


import os
import hashlib
import zlib


GIT_DIR = ".trackit"

def init():
    '''
    Generates the required file directory structure for TrackIt.
    '''
    os.makedirs(GIT_DIR, exist_ok=True)
    os.makedirs(f"{GIT_DIR}\\objects", exist_ok=True)
    os.makedirs(f"{GIT_DIR}\\refs", exist_ok=True)
    os.makedirs(f"{GIT_DIR}\\refs\\heads", exist_ok=True)
    os.makedirs(f"{GIT_DIR}\\refs\\tags", exist_ok=True)
    set_head(ref = 'refs\\heads\\main')
    

def hash_object(content, obj_type="blob"):
    '''
    Hash the object content and store it in the .trackit/objects.\n
    Returns Object ID.
    '''

    # '\0' is used to divide the header with content. Again we then encode the header to binary
    header = f"{obj_type} {len(content)}\0".encode()

    # content is already in binary format
    headed_content = header + content
    o_id = hashlib.sha1(headed_content).hexdigest()     #hashes using sha1 to get the object id

    # we divide the files with respect to their oid with first 2 characters as directory and the rest as file name
    obj_dir_path = os.path.join(GIT_DIR, 'objects', o_id[:2])
    os.makedirs(obj_dir_path, exist_ok = True)

    # compress the file to save storage
    with open(os.path.join(obj_dir_path, o_id[2:]), 'wb') as f:
        f.write(zlib.compress(headed_content))
    
    return o_id

def get_object(o_id, expected = 'blob'):
    '''
    Retrieves the object content with reference to the object id.\n
    Returns Binary Codec.
    '''
    try:
        # we first get the file, then we decompress it and then get the header and content using '\0' as reference.
        # now we decode the header to original format

        obj_path = os.path.join(GIT_DIR, 'objects', o_id[:2], o_id[2:])
        
        with open(obj_path, 'rb') as f:
            r = f.read()
        
        header, content = zlib.decompress(r).split(b'\0', 1)
        content_type = header.decode().split(" ")[0]
        content_size = header.decode().split(" ")[1]

        if expected:
            if expected != content_type: return f'Expected {expected}, got {content_type}'
        # print(content)
        return content
    
    except FileNotFoundError:
        print(f"Object {o_id} not found.")
        return None
    
def set_head(ref = None, o_id = None):
    '''
    Set the head to a commit.
    '''
    with open(os.path.join(GIT_DIR, 'HEAD'), 'w') as f:
        if ref:
            f.write(f"ref: {ref}")
        else:
            f.write(f"commit: {o_id}")

def get_head_branch():
    '''
    Gets the branch or commit ID to which the head is pointing to.\n
    Use check_symbolic to know if head is detached or not.
    '''
    with open(os.path.join(GIT_DIR, 'HEAD'), 'r') as f:
        return f.read().split(" ")[1]

def set_ref(ref, o_id):
    '''
    Set / Update the reference with to a commit.
    '''
    with open(os.path.join(GIT_DIR, ref), 'w') as f:
        f.write(o_id)

def get_ref(ref):
    '''
    Gets the commit ID of the reference.
    '''
    path = os.path.join(GIT_DIR, ref)

    if ref == 'HEAD':
        with open(path, 'r') as f:
            r = f.read().split(" ")
            if r[0] == 'commit:':
                return r[1]
            path = os.path.join(GIT_DIR, r[1])
    
    try:
        with open(path, 'r') as f:
            return f.read()
    except FileNotFoundError:
        return None

def iter_refs():
    '''
    Gets all the references (Branches and Tags) present in the current TrackIt repository.\n
    Returns ref, object id
    '''
    refs = ['HEAD', 'MERGE_HEAD']

    for root, dirs, files in os.walk(os.path.join(GIT_DIR, 'refs')):
        root = os.path.relpath(root, GIT_DIR)                   # this is done in order to remove GIT_DIR part from the path
        refs.extend([os.path.join(root, i) for i in files])
    
    for ref in refs:
        yield ref, get_ref(ref)

def check_symbolic():
    '''
    Check if the current head is detached or not.\n
    '''
    with open(os.path.join(GIT_DIR, 'HEAD'), 'r') as f:
        r = f.read()
        if r.split(" ")[0] == 'commit:':
            return False
        return True